{==============================================================================}
{
  Mail merge application ("GeoReport): editing template and perform merging.

  How it works:
  - RichViewEdit1 and RichView2 are linked to the same RVStyle (RVStyle1)
  - Template is edited in RichViewEdit1
  - RichViewEdit1 is scanned for fields. When field is found, it is deleted,
    and its value is inserted in its place:
    * field value is loaded in invisible RichView2 (styles can be added in RVStyle2)
    * contents of RichView2 is inserted in the proper place of RichViewEdit1.
}
{==============================================================================}

{
 This demo uses the most universal way for doing mail merging: fields are
 arbitrary text fragments (both in the mode when field codes are shown, and in
 when displaying field value)

 In this demo, a field is a set of adjacent items (of any type) marked with the
 same code. Field code is stored in items tags, in the form "Code|" for normal
 items, or "Code|Target" for hyperlinks.
 This representation has the following advantage: you can apply another set
 of field values without reloading a template.

 Mail merging is performed in RichViewEdit1. A hidden RichView2 is used
 for loading field data. Important: rvfoSaveTextStyles and rvfoSaveParaStyles
 must be excluded from RichView2.RVFOptions.

 This demo supports two types of field values:
 1) arbitrary document stored in RVF format in TBlobField
 2) a single line text stored in TStringField; for such fields, text attributes
    are defined by the text style returned by GetTextStyleForField procedure.


 History:
 Version 1.3 - initial release for FireDAC with Interbase tables
}


unit ReportFrm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Db, ExtCtrls, ShellApi,
  RVScroll, RichView, RVEdit, RVStyle, RVTypes, RVStrFuncs,
  Menus, CRVFData, RVTable, RVItem, ComCtrls, RVNormalize, CRVData, RVRVData,
  FireDAC.UI.Intf, FireDAC.VCLUI.Wait, FireDAC.Stan.Intf,
  FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.Phys.Intf, FireDAC.Stan.Def,
  FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.Param,
  FireDAC.DatS, FireDAC.DApt.Intf, FireDAC.DApt, FireDAC.Comp.DataSet,
  FireDAC.Comp.Client, FireDAC.Comp.UI, FireDAC.Phys.IBBase, FireDAC.Phys.IB;

type
  // Kinds of text that can be used for fields in this demo
  TDemoTextStyle = (dtsNormal, dtsHidden, dtsField);

type
  TForm1 = class(TForm)
    RichViewEdit1: TRichViewEdit;
    Panel1: TPanel;
    Label1: TLabel;
    ComboBox1: TComboBox;
    RVStyle1: TRVStyle;
    RichView2: TRichView;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Insert1: TMenuItem;
    Country1: TMenuItem;
    Capital1: TMenuItem;
    CallingCode1: TMenuItem;
    Description1: TMenuItem;
    Cities1: TMenuItem;
    Demographics1: TMenuItem;
    Exit1: TMenuItem;
    StatusBar1: TStatusBar;
    Edit1: TMenuItem;
    ClearCodesfromSelection1: TMenuItem;
    cbTableGrid: TCheckBox;
    FDPhysIBDriverLink1: TFDPhysIBDriverLink;
    FDGUIxWaitCursor1: TFDGUIxWaitCursor;
    FDConnection1: TFDConnection;
    FDTable1: TFDTable;
    procedure FormCreate(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure Country1Click(Sender: TObject);
    procedure Capital1Click(Sender: TObject);
    procedure CallingCode1Click(Sender: TObject);
    procedure Description1Click(Sender: TObject);
    procedure Cities1Click(Sender: TObject);
    procedure Demographics1Click(Sender: TObject);
    procedure ComboBox1Click(Sender: TObject);
    procedure RichViewEdit1CaretMove(Sender: TObject);
    procedure RichViewEdit1Jump(Sender: TObject; id: Integer);
    procedure ClearCodesfromSelection1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure cbTableGridClick(Sender: TObject);
  private
    { Private declarations }
    FDisplayCodes: Boolean; // Display field codes / field values
    procedure LoadCountryList(Strings: TStrings);
    procedure ReplaceFields(RVData: TCustomRVFormattedData);
    function LoadData(const Code: TRVUnicodeString; RVData: TCustomRVFormattedData;
      ItemNo: Integer): TMemoryStream;
    function GetTextStyleForField(const Code: TRVUnicodeString;
      RVData: TCustomRVFormattedData; ItemNo: Integer): Integer;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure ApplyCode(const Code: TRVUnicodeString; rv: TCustomRichView); forward;
{=========== PROCEDURES THAT DEPEND ON THE SET OF FIELDS ======================}
{ Returns an index of text style for the given kind of text. }
function GetTextStyle(Kind: TDemoTextStyle; RVStyle: TRVStyle; SourceStyleNo: Integer;
  RVData: TCustomRVFormattedData; ItemNo: Integer): Integer;
var Color, BackColor: TColor;
    TextStyle: TFontInfo;
begin
  if Kind = dtsField then
  begin
    Color := clYellow;
    BackColor := clBlack;
  end
  else
  begin
    Color := clWindowText;
    BackColor := clNone;
    if RVData<>nil then
    begin
      if (ItemNo>0) and not RVData.IsFromNewLine(ItemNo) and (RVData.GetItemStyle(ItemNo-1)>=0) then
      begin
        Color := RVData.GetRVStyle.TextStyles[RVData.GetItemStyle(ItemNo-1)].Color;
        BackColor := RVData.GetRVStyle.TextStyles[RVData.GetItemStyle(ItemNo-1)].BackColor;
      end
      else if (ItemNo+1<RVData.ItemCount) and not RVData.IsFromNewLine(ItemNo+1) and
        (RVData.GetItemStyle(ItemNo+1)>=0) then
      begin
        Color := RVData.GetRVStyle.TextStyles[RVData.GetItemStyle(ItemNo+1)].Color;
        BackColor := RVData.GetRVStyle.TextStyles[RVData.GetItemStyle(ItemNo+1)].BackColor;
      end;
    end;
  end;
  TextStyle := TFontInfo.Create(nil);
  try
    TextStyle.Assign(RVStyle.TextStyles[SourceStyleNo]);
    if Kind=dtsHidden then
      TextStyle.Options := TextStyle.Options+[rvteoHidden]
    else
      TextStyle.Options := TextStyle.Options-[rvteoHidden];
    TextStyle.Color := Color;
    TextStyle.BackColor := BackColor;
    Result := RVStyle.FindTextStyle(TextStyle);
  finally
    TextStyle.Free;
  end;
end;
{------------------------------------------------------------------------------}
{ Returns an index of text style to use for values of fields
  with the specific Code. }
function TForm1.GetTextStyleForField(const Code: TRVUnicodeString;
  RVData: TCustomRVFormattedData; ItemNo: Integer): Integer;
begin
  Result := GetTextStyle(dtsNormal, RVData.GetRVStyle, RVData.GetItemStyle(ItemNo),
    RVData, ItemNo);
end;
{------------------------------------------------------------------------------}
// Loading field code. You can implement your own ways for loading data:
// load them in RichView2
function TForm1.LoadData(const Code: TRVUnicodeString; RVData: TCustomRVFormattedData;
  ItemNo: Integer): TMemoryStream;
var
  Kind: TDemoTextStyle;
begin
  RichView2.Clear;
  Result := TMemoryStream.Create;
  try
    // Loading field value in RichView2.
    // Important: do not change existing styles in RichView2.Style,
    // because they are used in RichViewEdit1. Use InsertRVFFromStream, not
    // LoadRVFFromStream.
    if not FDisplayCodes then
    begin
      if FDTable1.FieldByName(Code) is TStringField then
        RichView2.AddNL(FDTable1.FieldByName(Code).AsString,
          GetTextStyleForField(Code, RVData, ItemNo), 0)
      else if FDTable1.FieldByName(Code) is TBlobField then
      begin
        TBlobField(FDTable1.FieldByName(Code)).SaveToStream(Result);
        Result.Position := 0;
        RichView2.InsertRVFFromStream(Result, 0); // inserting will merge styles;
               // RichViewEdit1 and RichView2 have the same collections of styles
      end
      else
        raise Exception.Create('Unsupported field type');
      Kind := dtsHidden
      end
    else
      Kind := dtsField;
    // If field value is empty, adding field code as a hidden text
    // (to prevent losing field)
    if (RichView2.ItemCount=0) or
      (RichView2.ItemCount=1) and (RichView2.GetItemStyle(0)>=0) and
      (RichView2.GetItemText(0)='') then
    begin
      RichView2.Clear;
      RichView2.AddNL(Code,
        GetTextStyle(Kind, RichView2.Style, RVData.GetItemStyle(ItemNo),
          RVData, ItemNo), 0);
    end;
    // Adding "Code|" to tags in RichView2.
    ApplyCode(Code, RichView2);
    // Saving RichView2 to the resulting stream
    Result.Clear;
    RichView2.SaveRVFToStream(Result, False);
    Result.Position := 0;
  except
    Result.Free;
    Result := nil;
  end;
end;
{==============================================================================}
{====== PROCEDURES THAT CAN BE INCLUDED IN YOUR APPLICATION UNCHANGED =========}
{==============================================================================}
{ Inserting field with the specified Code in the caret position of rve.
  The field has a text style returned by GetTextStyle(dtsField...). }
procedure InsertField(const Code: TRVUnicodeString; rve: TCustomRichViewEdit);
var StyleNo: Integer;
begin
  StyleNo := rve.CurTextStyleNo;
  rve.CurTextStyleNo := GetTextStyle(dtsField, rve.Style, StyleNo, nil, -1);
  rve.InsertStringWTag(Code, Code+'|');
  rve.CurTextStyleNo := StyleNo;
end;
{------------------------------------------------------------------------------}
{ Replacing fields with values. This procedure sometimes may create some
  errors in document, so calling NormalizeRichView is necessary after it.
  Besides, call ClearUndo method. }
procedure TForm1.ReplaceFields(RVData: TCustomRVFormattedData);
var
  i,j: Integer;
  Stream: TMemoryStream;
  Dummy1: TColor;
   Dummy2, Dummy3: Pointer;
   ParaNo: Integer;
   BR, ContinuePara: Boolean;
   table: TRVTableItemInfo;
   p, r,c: Integer;
   Code: TRVUnicodeString;
begin
  Dummy1 := clNone;
  Dummy2 := nil;
  Dummy3 := nil;
  i := RVData.ItemCount-1;
  while i>=0 do
  begin
    Code := PRVUnicodeChar(RVData.GetItemTag(i));
    p := RVGet1Index(RVPosW('|', Code));
    if p>0 then
    begin
      Code := Copy(Code, 1, p-1);
      // finding a range of items to delete (from i to j)
      j := i;
      while (i>=0) and (Copy(RVData.GetItemTag(i), 1, p-1)=Code) do
        dec(i);
      inc(i);
      // storing parameters of deleted items
      ParaNo := RVData.GetItemPara(i);
      BR     := RVData.GetItem(i).BR;
      ContinuePara := RVData.GetItem(i).SameAsPrev;
      // loading field value in the stream
      Stream := LoadData(Code, RVData, i);
      if Stream<>nil then
      begin
        // deleting the field code
        RVData.DeleteItems(i,j-i+1);
        // inserting the field value
        RVData.InsertRVFFromStream(Stream, i, Dummy1, Dummy2, Dummy3, False,
          nil, nil);
        // applying stored parameters to the inserted items
        if (i>0) and (RVData.GetItemStyle(i-1)=rvsListMarker) and
          (RVData.GetItemStyle(i)=rvsListMarker) then
          RVData.DeleteItems(i, 1);
        for j := i to RVData.ItemCount-1 do
        begin
          if (i=j) then
          begin
            if RVData.GetItem(j).GetBoolValue(rvbpFullWidth) then
            begin
              if (i>0) and (RVData.GetItemStyle(i-1)=rvsListMarker) then
              begin
                dec(i);
                RVData.DeleteItems(i, 1);
              end;
              break;
            end;
            RVData.GetItem(j).SameAsPrev := ContinuePara;
            if BR then
              RVData.GetItem(j).BR := True;
          end;
          if (j>i) and RVData.IsParaStart(j) then
            break;
          RVData.GetItem(j).ParaNo := ParaNo;
        end;
        Stream.Free;
      end;
    end
    else if RVData.GetItem(i) is TRVTableItemInfo then
    begin
      // recursive call for table cells
      table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to table.RowCount-1 do
        for c := 0 to table.ColCount-1 do
          if table.Cells[r,c]<>nil then
            ReplaceFields(table.Cells[r,c]);
    end;
    dec(i);
  end;
end;
{------------------------------------------------------------------------------}
{ Assigning tags (Code+'|') to all items in rv.
  If some items already have tags (for example, hyperlinks), this value is
  added to the beginning of existing tags. }
procedure ApplyCode(const Code: TRVUnicodeString; rv: TCustomRichView);
var
  i: Integer;
  s: TRVUnicodeString;
begin
  for i := 0 to rv.ItemCount-1 do
  begin
    s := Code+'|'+rv.GetItemTag(i);
    rv.SetItemTag(i, s);
  end;
end;
{------------------------------------------------------------------------------}
{ Returns a range of items containing selection or caret }
procedure GetSelectionBounds(rve: TCustomRichViewEdit;
  var StartNo, EndNo: Integer);
var StartOffs, EndOffs: Integer;
begin
   rve.RVData.GetSelectionBoundsEx(StartNo, StartOffs, EndNo, EndOffs, True);
   if (StartNo<>EndNo) or (StartOffs<>EndOffs) then
   begin
     if StartOffs>=rve.GetOffsAfterItem(StartNo) then
       inc(StartNo);
     if EndOffs<=rve.GetOffsBeforeItem(EndNo) then
       dec(EndNo);
   end;
end;
{------------------------------------------------------------------------------}
{ Removes field codes from tags of selected items (as undoable editing operation).
  For tags having the form "a|b", this procedure removes "a|" leaving only "b" }
procedure RemoveCodesFromSelection(rve: TCustomRichViewEdit);
var
  i, p, StartNo, EndNo: Integer;
  s: TRVUnicodeString;
begin
  rve := rve.TopLevelEditor;
  GetSelectionBounds(rve, StartNo, EndNo);
  if EndNo<StartNo then
    exit;
  rve.BeginUndoGroup(rvutTag);
  rve.SetUndoGroupMode(True);
  for i := StartNo to EndNo do
  begin
    s := rve.GetItemTag(i);
    p := RVGet1Index(RVPosW('|', s));
    if p>0 then
      s := Copy(s, p+1, MaxInt);
    if s<>'' then
      rve.SetItemTagEd(i, s)
    else
      rve.SetItemTagEd(i, '')
  end;
  rve.SetUndoGroupMode(False);
  rve.SetSelectionBounds(StartNo, rve.GetOffsBeforeItem(StartNo),
    EndNo, rve.GetOffsAfterItem(EndNo));
  rve.Invalidate;
end;
{==============================================================================}
{===================== PROCEDURES SPECIFIC FOR THIS DEMO ======================}
{==============================================================================}
procedure TForm1.FormCreate(Sender: TObject);
begin
  // loading document
  RichViewEdit1.Clear;
  if FileExists(ExtractFilePath(Application.ExeName)+'georeport.rvf') then
    RichViewEdit1.LoadRVF(ExtractFilePath(Application.ExeName)+'georeport.rvf');
  RichViewEdit1.Format;
  // opening table
  with FDConnection1.Params do
  begin
    Clear;
    Add('DriverID=IB');
    Add('Database=localhost:' + ExtractFilePath(Application.ExeName) + 'Geo.gdb');
    Add('User_Name=SYSDBA');
    Add('Password=masterkey');
  end;
  FDConnection1.Connected := True;
  FDTable1.Open('');
  // filling list of countries
  ComboBox1.Items.Clear;
  ComboBox1.Items.Add('(Field Codes');
  LoadCountryList(ComboBox1.Items);
  ComboBox1.ItemIndex := 0;
  // important settings
  RichView2.RVFOptions := RichView2.RVFOptions-[rvfoSaveTextStyles,
    rvfoSaveParaStyles];
end;
{------------------------------------------------------------------------------}
{ Adding list of countries to Strings }
procedure TForm1.LoadCountryList(Strings: TStrings);
begin
  FDTable1.First;
  while not FDTable1.Eof do begin
    Strings.Add(FDTable1.FieldByName('Country').AsString);
    FDTable1.Next;
  end;
end;
{------------------------------------------------------------------------------}
procedure TForm1.Exit1Click(Sender: TObject);
begin
  Close;
end;
{------------------------------------------------------------------------------}
procedure TForm1.Country1Click(Sender: TObject);
begin
  InsertField('Country', RichViewEdit1);
end;

procedure TForm1.Capital1Click(Sender: TObject);
begin
  InsertField('Capital', RichViewEdit1);
end;

procedure TForm1.CallingCode1Click(Sender: TObject);
begin
  InsertField('CallingCode', RichViewEdit1);
end;

procedure TForm1.Description1Click(Sender: TObject);
begin
  InsertField('Description', RichViewEdit1);
end;

procedure TForm1.Cities1Click(Sender: TObject);
begin
  InsertField('Cities', RichViewEdit1);
end;

procedure TForm1.Demographics1Click(Sender: TObject);
begin
  InsertField('Demographics', RichViewEdit1);
end;
{------------------------------------------------------------------------------}
{ Combobox click: do mail merging }
procedure TForm1.ComboBox1Click(Sender: TObject);
begin
  FDisplayCodes := ComboBox1.ItemIndex<=0;
  if not FDisplayCodes then
    FDTable1.RecNo := ComboBox1.ItemIndex;
  ReplaceFields(RichViewEdit1.RVData);
  NormalizeRichView(RichViewEdit1.RVData);
  RichViewEdit1.ClearUndo;
  RichViewEdit1.Format;
  RichViewEdit1.Modified := True; // to activate saving request on form closing
  RichViewEdit1.SetFocus;
end;
{------------------------------------------------------------------------------}
{ On caret moving: displaying where the caret is, in the status bar.
  Note: if the caret is in table cell, this function will show field only
  of the field is inside this table cell }
procedure TForm1.RichViewEdit1CaretMove(Sender: TObject);
var
  Code: TRVUnicodeString;
  p: Integer;
begin
  if RichViewEdit1.TopLevelEditor.CurItemNo<0 then
  begin
    StatusBar1.SimpleText := '';
    exit;
  end;
  Code := RichViewEdit1.TopLevelEditor.GetItemTag(
    RichViewEdit1.TopLevelEditor.CurItemNo);
  p := RVGet1Index(RVPosW('|', Code));
  if p>0 then
    Code := Copy(Code, 1, p-1)
  else
    Code := '';
  if Code='' then
    StatusBar1.SimpleText := 'The caret is not in any field'
  else
    StatusBar1.SimpleText := 'The caret is in field: '+Code;
end;
{------------------------------------------------------------------------------}
{ Clearing codes from the selected text }
procedure TForm1.ClearCodesfromSelection1Click(Sender: TObject);
begin
  RemoveCodesFromSelection(RichViewEdit1);
  RichViewEdit1CaretMove(RichViewEdit1);
end;
{------------------------------------------------------------------------------}
{ On Ctrl+clicking hyperlink: extracting non-code part of tag ("b" from "a|b")
  and opening it as a hyperlink target or a file }
procedure TForm1.RichViewEdit1Jump(Sender: TObject; id: Integer);
var
  RVData: TCustomRVFormattedData;
  p,ItemNo: Integer;
  Tag: TRVTag;
begin
  TCustomRichView(Sender).GetJumpPointLocation(id, RVData, ItemNo);
  Tag := RVData.GetItemTag(ItemNo);
  p := RVGet1Index(RVPosW('|', Tag));
  if p>0 then
    Tag := Copy(Tag, p+1, MaxInt);
  ShellExecuteW(0, 'open', PRVUnicodeChar(Tag), nil, nil, SW_SHOW);
end;
{------------------------------------------------------------------------------}
{ Saving document on closing }
procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if not RichViewEdit1.Modified then
    exit;

  case Application.MessageBox('Template was modified, save?', 'Closing',
    MB_YESNOCANCEL or MB_ICONQUESTION) of
    IDYES:
      begin
        RichViewEdit1.DeleteUnusedStyles(True, True, True);
        RichViewEdit1.SaveRVF(ExtractFilePath(Application.ExeName)+'georeport.rvf', False);
        CanClose := True;
      end;
    IDNO:
      CanClose := True;
    IDCANCEL:
      CanClose := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TForm1.cbTableGridClick(Sender: TObject);
begin
  if cbTableGrid.Checked then
    RichViewEdit1.Options := RichViewEdit1.Options + [rvoShowGridLines]
  else
    RichViewEdit1.Options := RichViewEdit1.Options - [rvoShowGridLines];
  RichViewEdit1.Reformat;
end;

end.